"""
Molecular integrals over Gaussian basis functions generated by sympleints.
See https://github.com/eljost/sympleints for more information.

sympleints version: 0.1.dev79+g63f1ef8.d20230515
symppy version: 1.10.1

sympleints was executed with the following arguments:
	lmax = 4
	lauxmax = 6
	write = False
	out_dir = devel_ints
	keys = ['~2c2e', '~3c2e_sph']
	sph = False
	opt_basic = True
	normalize = cgto
"""

import numpy


def cart_gto3d_0(ax, da, A, R):
    """3D Cartesian s-Gaussian shell.
    Exponent ax, contraction coeff. da, centered at A, evaluated at R.

    Generated code; DO NOT modify by hand!"""

    result = numpy.zeros((1,), dtype=float)

    # 1 item(s)
    result[0] = numpy.sum(
        da
        * numpy.exp(-ax * ((A[0] - R[0]) ** 2 + (A[1] - R[1]) ** 2 + (A[2] - R[2]) ** 2))
    )
    return result


def cart_gto3d_1(ax, da, A, R):
    """3D Cartesian p-Gaussian shell.
    Exponent ax, contraction coeff. da, centered at A, evaluated at R.

    Generated code; DO NOT modify by hand!"""

    result = numpy.zeros((3,), dtype=float)

    x0 = A[0] - R[0]
    x1 = A[1] - R[1]
    x2 = A[2] - R[2]
    x3 = da * numpy.exp(-ax * (x0**2 + x1**2 + x2**2))

    # 3 item(s)
    result[0] = numpy.sum(-x0 * x3)
    result[1] = numpy.sum(-x1 * x3)
    result[2] = numpy.sum(-x2 * x3)
    return result


def cart_gto3d_2(ax, da, A, R):
    """3D Cartesian d-Gaussian shell.
    Exponent ax, contraction coeff. da, centered at A, evaluated at R.

    Generated code; DO NOT modify by hand!"""

    result = numpy.zeros((6,), dtype=float)

    x0 = A[0] - R[0]
    x1 = x0**2
    x2 = A[1] - R[1]
    x3 = x2**2
    x4 = A[2] - R[2]
    x5 = x4**2
    x6 = da * numpy.exp(-ax * (x1 + x3 + x5))
    x7 = 0.5773502691896258 * x6
    x8 = x0 * x6

    # 6 item(s)
    result[0] = numpy.sum(x1 * x7)
    result[1] = numpy.sum(x2 * x8)
    result[2] = numpy.sum(x4 * x8)
    result[3] = numpy.sum(x3 * x7)
    result[4] = numpy.sum(x2 * x4 * x6)
    result[5] = numpy.sum(x5 * x7)
    return result


def cart_gto3d_3(ax, da, A, R):
    """3D Cartesian f-Gaussian shell.
    Exponent ax, contraction coeff. da, centered at A, evaluated at R.

    Generated code; DO NOT modify by hand!"""

    result = numpy.zeros((10,), dtype=float)

    x0 = A[0] - R[0]
    x1 = x0**2
    x2 = A[1] - R[1]
    x3 = x2**2
    x4 = A[2] - R[2]
    x5 = x4**2
    x6 = da * numpy.exp(-ax * (x1 + x3 + x5))
    x7 = 0.2581988897471611 * x6
    x8 = x2 * x6
    x9 = 0.5773502691896258
    x10 = x1 * x9
    x11 = x4 * x6
    x12 = x3 * x9
    x13 = x0 * x6
    x14 = x5 * x9

    # 10 item(s)
    result[0] = numpy.sum(-(x0**3) * x7)
    result[1] = numpy.sum(-x10 * x8)
    result[2] = numpy.sum(-x10 * x11)
    result[3] = numpy.sum(-x12 * x13)
    result[4] = numpy.sum(-x0 * x4 * x8)
    result[5] = numpy.sum(-x13 * x14)
    result[6] = numpy.sum(-(x2**3) * x7)
    result[7] = numpy.sum(-x11 * x12)
    result[8] = numpy.sum(-x14 * x8)
    result[9] = numpy.sum(-(x4**3) * x7)
    return result


def cart_gto3d_4(ax, da, A, R):
    """3D Cartesian g-Gaussian shell.
    Exponent ax, contraction coeff. da, centered at A, evaluated at R.

    Generated code; DO NOT modify by hand!"""

    result = numpy.zeros((15,), dtype=float)

    x0 = -A[0] + R[0]
    x1 = x0**2
    x2 = -A[1] + R[1]
    x3 = x2**2
    x4 = -A[2] + R[2]
    x5 = x4**2
    x6 = da * numpy.exp(-ax * (x1 + x3 + x5))
    x7 = 0.09759000729485332 * x6
    x8 = 0.2581988897471611 * x6
    x9 = x0**3 * x8
    x10 = 0.3333333333333333 * x6
    x11 = x1 * x10
    x12 = 1.732050807568877
    x13 = x12 * x4
    x14 = x2**3
    x15 = x0 * x8
    x16 = x10 * x3
    x17 = x4**3

    # 15 item(s)
    result[0] = numpy.sum(x0**4 * x7)
    result[1] = numpy.sum(x2 * x9)
    result[2] = numpy.sum(x4 * x9)
    result[3] = numpy.sum(x11 * x3)
    result[4] = numpy.sum(x11 * x13 * x2)
    result[5] = numpy.sum(x11 * x5)
    result[6] = numpy.sum(x14 * x15)
    result[7] = numpy.sum(x0 * x13 * x16)
    result[8] = numpy.sum(x0 * x10 * x12 * x2 * x5)
    result[9] = numpy.sum(x15 * x17)
    result[10] = numpy.sum(x2**4 * x7)
    result[11] = numpy.sum(x14 * x4 * x8)
    result[12] = numpy.sum(x16 * x5)
    result[13] = numpy.sum(x17 * x2 * x8)
    result[14] = numpy.sum(x4**4 * x7)
    return result


cart_gto3d = {
    (0,): cart_gto3d_0,
    (1,): cart_gto3d_1,
    (2,): cart_gto3d_2,
    (3,): cart_gto3d_3,
    (4,): cart_gto3d_4,
}



# https://verahill.blogspot.de/2013/06/439-calculate-frequencies-from-hessian.html
# https://chemistry.stackexchange.com/questions/74639

import logging
from math import ceil, log
import os
from pathlib import Path
import sys

import h5py
import numpy as np

from pysisyphus.constants import BOHR2ANG, AU2KJPERMOL
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers import check_for_end_sign
from pysisyphus.helpers_pure import (
    highlight_text,
    eigval_to_wavenumber,
    report_isotopes,
    rms,
)
from pysisyphus.irc.Instanton import T_crossover_from_eigval
from pysisyphus.io import save_third_deriv
from pysisyphus.optimizers.guess_hessians import get_guess_hessian
from pysisyphus.TablePrinter import TablePrinter
from pysisyphus.xyzloader import make_trj_str, make_xyz_str

import torch


class IRC:
    valid_displs = ("energy", "length")

    def __init__(
        self,
        geometry,
        step_length=0.1,
        max_cycles=125,
        downhill=False,
        forward=True,
        backward=True,
        root=0,
        hessian_init=None,
        displ="energy",
        displ_energy=1e-3,
        displ_length=0.1,
        rms_grad_thresh=1e-3,
        hard_rms_grad_thresh=None,
        energy_thresh=1e-6,
        imag_below=0.0,
        force_inflection=True,
        check_bonds=False,
        out_dir=".",
        prefix="",
        dump_fn="irc_data.h5",
        dump_every=5,
    ):
        """Base class for IRC calculations.

        Parameters
        ----------
        geometry : Geometry
            Transtion state geometry, or initial geometry for downhill run.
        step_length : float, optional
            Step length in unweighted coordinates.
        max_cycles : int, optional
            Positive integer, controlloing the maximum number of IRC steps
            taken in a direction (forward/backward/downhill).
        downhill : bool, default=False
            Downhill run from a non-stationary point with non-vanishing
            gradient. Disables forward and backward runs.
        forward : bool, default=True
            Integrate IRC in positive s direction.
        backward : bool, default=True
            Integrate IRC in negative s direction.
        root : int, default=0
            Use n-th root for initial displacement from TS.
        hessian_init : str, default=None
            Path to Hessian HDF5 file, e.g., from a previous TS calculation.
        displ: str, one of ("energy", "length")
            Controlls initial displacement from the TS. 'energy' assumes a
            quadratic model, from which a step length for a given energy
            lowering (see 'displ_energy') is determined. 'length' corresponds
            to a displacement along the transition vector. 
        displ_energy : float, default=1e-3
            Required energy lowering from the TS in au (Hartree). Used with
            'displ: energy'.
        displ_length : float, default=0.1
            Step length along the transition vector. Used only with
            'displ: length'.
        rms_grad_thresh : float, default=1e-3,
            Convergence is signalled when to root mean square of the unweighted
            gradient is less than or equal to this value
        energy_thresh : float, default=1e-6,
            Signal convergence when the energy difference between two points
            is equal to or less than 'energy_thresh'.
        imag_below : float, default=0.0
            Require the wavenumber of the imaginary mode to be below the
            given threshold. If given, it should be a negative number.
        force_inflection : bool, optional
            Don't indicate convergence before passing an inflection point.
        check_bonds : bool, optional, default=True
            Report whether bonds are formed/broken along the IRC, w.r.t the TS.
        out_dir : str, optional
            Dump everything into 'out_dir' directory instead of the CWD.
        prefix : str, optional
            Short string that is prepended to all files that are created
            by this class, e.g., trajectories and HDF5 dumps.
        dump_fn : str, optional
            Base name for the HDF5 files.
        dump_every : int, optional
            Dump to HDF5 every n-th cycle.
        """
        assert step_length > 0, "step_length must be positive"
        assert max_cycles > 0, "max_cycles must be positive"

        self.logger = logging.getLogger("irc")

        self.geometry = geometry
        self.atoms = self.geometry.atoms
        assert self.geometry.coord_type == "cart"

        report_isotopes(self.geometry, "the IRC")

        self.step_length = step_length
        self.max_cycles = max_cycles
        self.downhill = downhill
        # Disable forward/backward when downhill is set
        self.forward = not self.downhill and forward
        self.backward = not self.downhill and backward
        self.root = root
        if hessian_init is None:
            hessian_init = "calc" if not self.downhill else "unit"
        self.hessian_init = hessian_init
        self.displ = displ
        assert (
            self.displ in self.valid_displs
        ), f"'displ: {self.displ}' not in {self.valid_displs}"
        self.displ_energy = float(displ_energy)
        self.displ_length = float(displ_length)
        self.rms_grad_thresh = float(rms_grad_thresh)
        self.hard_rms_grad_thresh = hard_rms_grad_thresh
        self.energy_thresh = float(energy_thresh)
        assert imag_below <= 0.0
        self.imag_below = imag_below
        self.force_inflection = force_inflection
        self.check_bonds = check_bonds
        self.out_dir = out_dir
        self.out_dir = Path(self.out_dir)
        if not self.out_dir.exists():
            os.mkdir(self.out_dir)
        self.prefix = f"{prefix}_" if prefix else prefix
        self.dump_fn = dump_fn
        self.dump_every = int(dump_every)

        # Determine bonds at TS
        self.ts_bond_sets = self.geometry.bond_sets
        self.ref_bond_sets = {}

        self._m_sqrt = np.sqrt(self.geometry.masses_rep)

        # cache frequently‑used active indices
        if getattr(self.geometry, "within_partial_hessian", None) is not None:
            self._act_atoms = self.geometry.hess_active_atom_indices
            self._act_dofs = self.geometry.hess_active_dof_indices
        else:
            self._act_atoms = self.geometry.active_atom_indices
            self._act_dofs  = self.geometry.active_dof_indices

        self.all_energies = list()
        self.all_coords = list()
        self.all_gradients = list()
        self.all_mw_coords = list()
        self.all_mw_gradients = list()

        # step length dE max(|grad|) rms(grad)
        col_fmts = "int float float float float".split()
        header = ("Step", "IRC length", "dE / au", "max(|grad|)", "rms(grad)")
        self.table = TablePrinter(header, col_fmts)

        self.cycle_places = ceil(log(self.max_cycles, 10))

        self.mm_inv2 = self.geometry.mm_sqrt_inv[np.ix_(self._act_dofs, self._act_dofs)]

    def get_path_for_fn(self, fn):
        return self.out_dir / f"{self.prefix}{fn}"

    @property
    def coords(self):
        return self.geometry.coords

    @coords.setter
    def coords(self, coords):
        self.geometry.coords = coords

    @property
    def mw_coords(self):
        return self.geometry.mw_coords

    @mw_coords.setter
    def mw_coords(self, mw_coords):
        self.geometry.mw_coords = mw_coords

    @property
    def energy(self):
        return self.geometry.energy

    @property
    def gradient(self):
        return self.geometry.gradient

    @property
    def mw_gradient(self):
        return self.geometry.mw_gradient

    # @property
    # def mw_hessian(self):
    # # TODO: This can be removed when the mw_hessian property is updated
    # #       in Geometry.py.
    # return self.geometry.mw_hessian

    def log(self, msg):
        # self.logger.debug(f"step {self.cur_cycle:03d}, {msg}")
        self.logger.debug(msg)

    @property
    def m_sqrt(self):
        return self._m_sqrt

    def unweight_vec(self, vec):
        return self.m_sqrt * vec

    def mass_weigh_hessian(self, hessian):
        return self.geometry.mass_weigh_hessian(hessian)
    
    # mass‑weight only the sub‑Hessian that belongs to the moving atoms
    def _mw_hessian_active(self, H_act):
        if isinstance(H_act, torch.Tensor):
            if not isinstance(self.mm_inv2, torch.Tensor):
                self.mm_inv2 = torch.as_tensor(self.mm_inv2, dtype=H_act.dtype, device=H_act.device)
            return self.mm_inv2 @ H_act @ self.mm_inv2       # in‑place not possible → tiny matrix
        return self.mm_inv2.dot(H_act).dot(self.mm_inv2)
    
    # Eckart projector that *ignores* frozen atoms
    def _project_active(self, mw_H_act, *, return_P=False):
        if self.geometry.is_analytical_2d or mw_H_act.shape[0] <= 6:
            return (mw_H_act, None) if return_P else mw_H_act

        from pysisyphus.Geometry import get_trans_rot_projector

        coords_act  = self.geometry.coords3d[self._act_atoms].flatten()
        masses_act  = self.geometry.masses[self._act_atoms]
        P           = get_trans_rot_projector(coords_act, masses=masses_act, full=False)

        if isinstance(mw_H_act, torch.Tensor):
            P = torch.as_tensor(P, dtype=mw_H_act.dtype, device=mw_H_act.device)
            proj = (P @ mw_H_act @ P.T)
            proj = 0.5 * (proj + proj.T)   # restore symmetry
        else:
            proj = P.dot(mw_H_act).dot(P.T)
            proj = 0.5 * (proj + proj.T)
        return (proj, P) if return_P else proj

    # Expand an active‑vector (or ‑step) to 3N
    def _full(self, vec_act):
        inds = self._act_dofs
        if isinstance(vec_act, torch.Tensor):
            idx = torch.as_tensor(inds, dtype=torch.long, device=vec_act.device)
            full = torch.zeros(self.coords.size, dtype=vec_act.dtype, device=vec_act.device)
            full.index_copy_(0, idx, vec_act)
            return full
        full = np.zeros(self.coords.size, dtype=vec_act.dtype if hasattr(vec_act, "dtype") else float)
        full[inds] = vec_act
        return full

    def prepare(self, direction):
        self.direction = direction
        self.converged = False
        self.energy_increased = False
        self.energy_converged = False
        self.past_inflection = not self.force_inflection

        self.irc_energies = list()
        # Not mass-weighted
        self.irc_coords = list()
        self.irc_gradients = list()
        # Mass-weighted
        self.irc_mw_coords = list()
        self.irc_mw_gradients = list()

        self.ref_bond_sets = self.ts_bond_sets.copy()

        # Over the course of the IRC the hessian may get updated.
        # Copying the initial hessian here ensures a clean start in combined
        # forward and backward runs. Otherwise we may accidentally use
        # the updated hessian from the end of the first run for the second
        # run.

        self.mw_hessian = self._mw_hessian_active(self.init_hessian)

        trj_fn = self.get_path_for_fn(f"{direction}_irc_trj.xyz")
        self.trj_handle = open(trj_fn, "w")

        # We don't need an initial displacement when going downhill
        if self.downhill:
            return

        # Do inital displacement from the TS
        if direction == "forward":
            initial_step = self.init_displ_plus
        elif direction == "backward":
            initial_step = self.init_displ_minus
        else:
            raise Exception("Invalid direction='{direction}'!")
        self.coords = self.ts_coords + initial_step

        if self.displ in ("energy"):
            actual_energy = self.energy
            actual_lowering = self.ts_energy - actual_energy
            diff = self.displ_energy - actual_lowering

            def en_str(en):
                return f"{en: .4f} au ({en*AU2KJPERMOL: .2f} kJ mol⁻¹)"

            print(
                f"Requested energy lowering: {en_str(self.displ_energy)}\n"
                f"   Actual energy lowering: {en_str(actual_lowering)}\n"
                f"                        Δ: {en_str(diff)}"
            )
            if actual_lowering < 0.0:
                print("Displaced geometry is higher in energy. Bisecting step...")
                step = initial_step.copy()
                for _ in range(5):
                    step = step * 0.5
                    self.coords = self.ts_coords + step
                    actual_lowering = self.ts_energy - self.energy
                    if actual_lowering > 0.0:
                        print(f"  Resolved: lowering={actual_lowering:.6f} au")
                        break
                else:
                    print("WARNING: Could not find downhill displacement.")
            print("\n")
            sys.stdout.flush()
        initial_step_length = np.linalg.norm(initial_step)
        self.logger.info(
            f"Did inital step of length {initial_step_length:.4f} " "from the TS."
        )

    def initial_displacement(self):
        """Returns non-mass-weighted steps in +s and -s direction
        for initial displacement from the TS. Earlier version only
        returned one step, that was later multiplied by either 1 or -1,
        depending on the desired IRC direction (forward/backward).
        The current implementation directly returns two steps for forward
        and backward direction. Whereas for plus and minus steps for
        displ 'length' and displ 'energy'
            step_plus = -step_minus
        is valid. The latter step is formed as
            x(ds) = ds * v0 + ds**2 * v1
        so
            x(ds) != -x(ds)
        as
            ds * v0 + ds**2 * v1 != -ds * v0 - ds**2 * v1 .

        So, all required step are formed directly and later used as appropriate.

        See
            https://aip.scitation.org/doi/pdf/10.1063/1.454172
            https://pubs.acs.org/doi/10.1021/j100338a027
            https://aip.scitation.org/doi/pdf/10.1063/1.459634
        """

        mw_hessian = self._mw_hessian_active(self.init_hessian)
        if self.coords.size > 3:
            proj_hessian, P = self._project_active(mw_hessian, return_P=True)
        # Don't project single atom species and analytical potentials
        else:
            proj_hessian = mw_hessian
            if isinstance(proj_hessian, torch.Tensor):
                P = torch.eye(self.coords.size, device=proj_hessian.device, dtype=proj_hessian.dtype)
            else:
                P = np.eye(self.coords.size)

        del mw_hessian

        if isinstance(proj_hessian, torch.Tensor):
            eigvals, eigvecs = torch.linalg.eigh(proj_hessian)
            eigvals = eigvals.to(torch.double).cpu().numpy()
            mw_cart_displs = P.T @ eigvecs if P is not None else eigvecs
            if not isinstance(self.mm_inv2, torch.Tensor):
                self.mm_inv2 = torch.as_tensor(self.mm_inv2, dtype=proj_hessian.dtype, device=proj_hessian.device)
            cart_displs = self.mm_inv2 @ mw_cart_displs
        else:
            eigvals, eigvecs = np.linalg.eigh(proj_hessian)    
            mw_cart_displs = P.T.dot(eigvecs)
            cart_displs = self.mm_inv2.dot(mw_cart_displs)
        
        nus = eigval_to_wavenumber(eigvals)
        nu_root = nus[self.root]
        assert nu_root <= self.imag_below, (
            f"Wavenumber {nu_root:.2f} cm⁻¹ of imaginary mode {self.root} is above "
            f"the threshold of {self.imag_below:.2f} cm⁻¹."
        )
        neg_inds = eigvals < -1e-8
        assert sum(neg_inds) > 0, "The hessian does not have any negative eigenvalues!"

        min_eigval = eigvals[self.root]
        min_nu = nus[self.root]
        T_c = T_crossover_from_eigval(min_eigval)
        min_msg = (
            f"Transition vector is mode {self.root} with wavenumber {min_nu:.2f} cm⁻¹.\n"
            f"Crossover temperature T_c: {T_c:.2f} K"
        )
        # Doing it this way hurts ... I'll have to improve my logging game...
        self.log(min_msg)
        print(min_msg)

        # Mass-weighted
        mw_trans_vec = mw_cart_displs[:, self.root]
        if isinstance(mw_trans_vec, torch.Tensor):
            mw_trans_vec = mw_trans_vec.cpu().numpy()
        
        # Not mass-weighted
        trans_vec = cart_displs[:, self.root]
        if isinstance(trans_vec, torch.Tensor):
            trans_vec = trans_vec.cpu().numpy()

        mw_trans_vec = self._full(mw_trans_vec)
        trans_vec = self._full(trans_vec)

        self.mw_transition_vector = mw_trans_vec        
        self.transition_vector = trans_vec / np.linalg.norm(trans_vec)

        if self.downhill:
            mw_step_plus = mw_step_minus = np.zeros_like(self.transition_vector)
            msg = "Downhill run. No initial displacement from the TS."
        elif self.displ == "length":
            msg = "Using length-based initial displacement from the TS."
            mw_step_plus = self.displ_length * mw_trans_vec
            mw_step_minus = -mw_step_plus
        elif self.displ == "energy":
            # Calculate the length of the initial step away from the TS to initiate
            # the IRC/MEP. We assume a quadratic potential and calculate the
            # displacement for a given energy lowering.
            # dE = (k*dq**2)/2 (dE = energy lowering, k = eigenvalue corresponding
            # to the transition vector/imaginary mode, dq = step length)
            # dq = sqrt(dE*2/k)
            # See 10.1021/ja00295a002 and 10.1063/1.462674
            # 10.1002/jcc.540080808 proposes 3 kcal/mol as initial energy lowering
            msg = (
                f"Energy-based (ΔE={self.displ_energy} au) initial displacement from "
                "the TS using 2rd derivatives."
            )
            step_length = np.sqrt(self.displ_energy * 2 / np.abs(min_eigval))
            # Guard against near-zero eigenvalue producing an excessively
            # large initial displacement.
            max_displ = 3.0  # au in mass-weighted coordinates
            if step_length > max_displ:
                print(
                    f"Warning: energy-based initial displacement {step_length:.4f} au "
                    f"exceeds {max_displ} au (|eigval|={np.abs(min_eigval):.6e}). "
                    f"Clamping to {max_displ} au."
                )
                self.log(
                    f"Clamped initial displacement from {step_length:.4f} to "
                    f"{max_displ} au."
                )
                step_length = max_displ
            # This calculation is derived from the mass-weighted hessian, so we
            # have to multiply this step length with the mass-weighted
            # mode and un-weigh it.
            mw_step_plus = step_length * mw_trans_vec
            mw_step_minus = -mw_step_plus
        else:
            raise Exception(f"self.displ={self.displ} is invalid!")

        step_plus = mw_step_plus / self.m_sqrt
        step_minus = mw_step_minus / self.m_sqrt
        self.log(msg)
        print(msg)
        print(
            "Initial step lengths (not mass-weighted):\n"
            f"\t Forward: {np.linalg.norm(step_plus):.4f} au\n"
            f"\tBackward: {np.linalg.norm(step_minus):.4f} au"
        )
        return step_plus, step_minus

    def get_conv_fact(self, mw_grad, min_fact=2.0):
        # Numerical integration of differential equations requires a step length and/or
        # we have to terminate the integration at some point, e.g. when the desired
        # step length is reached. IRCs are integrated in mass-weighted coordinates,
        # but self.step_length is given in unweighted coordinates. Unweighting a step
        # in mass-weighted coordinates will reduce its norm as we divide by sqrt(m).
        #
        # If we want to do an Euler-integration we have to decide on a step size
        # when a desired integration length is to be reached in a given number of steps.
        # [3] proposes using Δs/250 with a maximum of 500 steps, so something like
        # Δs/(max_steps / 2). It seems we can't use this because (at
        # least for the systems I tested) this will lead to a step length that is too
        # small, so the predictor Euler-integration will fail to converge in the
        # prescribed number of cycles. It fails because simply dividing the desired
        # step length in unweighted coordinates does not take into account the mass
        # dependence. Such a step size is appropriate for integrations in unweighted
        # coordinates, but not when using mass-weighted coordinates.
        #
        # We determine a conversion factor from comparing the magnitudes (norms) of
        # the mass-weighted and un-mass-weighted gradients. This takes into account
        # which atoms are actually moving, so it should be a good guess.
        norm_mw_grad = np.linalg.norm(mw_grad)
        if mw_grad.shape[0] == self._m_sqrt.shape[0]:
            norm_grad = np.linalg.norm(self.unweight_vec(mw_grad))
        else:
            m_sqrt_vec = self._m_sqrt[self._act_dofs]
            norm_grad = np.linalg.norm(mw_grad * m_sqrt_vec)

        if not np.isfinite(norm_mw_grad) or norm_mw_grad == 0.0:
            conv_fact = min_fact
            self.log("mw_grad norm is zero/NaN; using minimum conversion factor.")
        else:
            conv_fact = norm_grad / norm_mw_grad
            # Cap conversion factor when using an active subspace to avoid huge steps.
            if mw_grad.shape[0] != self._m_sqrt.shape[0]:
                max_fact = 50.0
                if conv_fact > max_fact:
                    self.log(f"Clamping conversion factor {conv_fact:.4f} -> {max_fact:.1f}.")
                    conv_fact = max_fact
            conv_fact = max(min_fact, conv_fact)
        self.log(f"Un-weighted / mass-weighted conversion factor {conv_fact:.4f}")
        return conv_fact

    def report_bonds(self, prefix, bonds):
        if len(bonds) == 0:
            return

        plural = "s" if len(bonds) > 1 else ""
        bond_strs = list()
        for from_, to_ in bonds:
            from_atom = self.atoms[from_]
            to_atom = self.atoms[to_]
            bond_strs.append(f"[{from_atom}{from_}-{to_atom}{to_}]")
        bonds_str = ", ".join(bond_strs)
        self.table.print(f"Bond{plural} {prefix}: {bonds_str}")

    def irc(self, direction):
        self.log(highlight_text(f"IRC {direction}", level=1))
        self.cur_direction = direction
        self.prepare(direction)
        # Calculate gradient
        self.gradient
        self.irc_energies.append(self.energy)
        # Non mass-weighted
        self.irc_coords.append(self.coords)
        self.irc_gradients.append(self.gradient)
        # Mass-weighted
        self.irc_mw_coords.append(self.mw_coords)
        self.irc_mw_gradients.append(self.mw_gradient)

        self.table.print_header()
        for self.cur_cycle in range(self.max_cycles):
            self.log(highlight_text(f"IRC step {self.cur_cycle:03d}") + "\n")

            # Dump current coordinates to trj
            comment = f"{direction} IRC, step {self.cur_cycle}"
            coords_str = make_xyz_str(
                self.atoms, BOHR2ANG * self.coords.reshape((-1, 3)), comment
            )
            self.trj_handle.write(coords_str + "\n")
            self.trj_handle.flush()

            self.log(f"Current energy: {self.energy:.6f} au")
            #
            # Take IRC step.
            #
            self.step()

            # Calculate gradient and energy on the new geometry
            # Non mass-weighted
            self.log("Calculating energy and gradient at new geometry.")
            self.irc_coords.append(self.coords)
            self.irc_gradients.append(self.gradient)
            self.irc_energies.append(self.energy)
            # Mass-weighted
            self.irc_mw_coords.append(self.mw_coords)
            self.irc_mw_gradients.append(self.mw_gradient)

            rms_grad = rms(self.gradient)

            # Only update once
            if not self.past_inflection:
                self.past_inflection = rms_grad >= self.rms_grad_thresh
                _ = "" if self.past_inflection else "not yet"
                self.log(f"(rms(grad) > threshold) {_} fullfilled!")

            irc_length = np.linalg.norm(self.irc_mw_coords[0] - self.irc_mw_coords[-1])
            dE = self.irc_energies[-1] - self.irc_energies[-2]
            max_grad = np.abs(self.gradient).max()

            row_args = (self.cur_cycle, irc_length, dE, max_grad, rms_grad)
            self.table.print_row(row_args)
            try:
                # The derived IRC classes may want to do some printing
                add_info = self.get_additional_print()
                self.table.print(add_info)
            except AttributeError:
                pass

            if self.check_bonds:
                cur_bond_sets = self.geometry.bond_sets
                formed = cur_bond_sets - self.ref_bond_sets
                broken = self.ref_bond_sets - cur_bond_sets
                self.report_bonds("formed", formed)
                self.report_bonds("broken", broken)
                # Update bond sets to avoid repeated reporting of bond topology changes
                self.ref_bond_sets -= broken
                self.ref_bond_sets |= formed  # union

            last_energy = self.irc_energies[-2]
            this_energy = self.irc_energies[-1]

            break_msg = ""
            self.energy_increased = this_energy > last_energy
            self.energy_converged = abs(last_energy - this_energy) <= self.energy_thresh
            if self.converged:
                break_msg = "Integrator indicated convergence!"
            elif self.past_inflection and (rms_grad <= self.rms_grad_thresh):
                break_msg = "rms(grad) converged!"
                self.converged = True
            elif (
                self.hard_rms_grad_thresh
                and (not self.past_inflection)
                and (rms_grad <= self.hard_rms_grad_thresh)
            ):
                break_msg = "rms(grad) below hard threshold."
            # TODO: Allow some threshold?
            elif self.energy_increased:
                break_msg = "Energy increased!"
            elif self.energy_converged:
                break_msg = "Energy converged!"
                self.converged = True

            # dumped = (self.cur_cycle % self.dump_every) == 0
            # if dumped:
            #     dump_fn = self.get_path_for_fn(f"{direction}_{self.dump_fn}")
            #     self.dump_data(dump_fn)

            if break_msg:
                self.table.print(break_msg)
                break

            if check_for_end_sign():
                break
            self.log("")
            sys.stdout.flush()
        else:
            print("IRC steps exceeded. Stopping.")
            print()

        if direction == "forward":
            self.irc_energies.reverse()
            self.irc_coords.reverse()
            self.irc_gradients.reverse()
            self.irc_mw_coords.reverse()
            self.irc_mw_gradients.reverse()

        # if not dumped:
        #     self.dump_data(dump_fn)

        self.cur_direction = None
        self.trj_handle.close()

    def set_data(self, prefix):
        energies_name = f"{prefix}_energies"
        coords_name = f"{prefix}_coords"
        grad_name = f"{prefix}_gradients"
        mw_coords_name = f"{prefix}_mw_coords"
        mw_grad_name = f"{prefix}_mw_gradients"

        setattr(self, coords_name, self.irc_coords)
        setattr(self, grad_name, self.irc_gradients)
        setattr(self, mw_coords_name, self.irc_mw_coords)
        setattr(self, mw_grad_name, self.irc_mw_gradients)
        setattr(self, energies_name, self.irc_energies)

        self.all_energies.extend(getattr(self, energies_name))
        self.all_coords.extend(getattr(self, coords_name))
        self.all_gradients.extend(getattr(self, grad_name))
        self.all_mw_coords.extend(getattr(self, mw_coords_name))
        self.all_mw_gradients.extend(getattr(self, mw_grad_name))

        # Free per-direction lists to reduce memory usage
        del self.irc_coords
        del self.irc_gradients
        del self.irc_mw_coords
        del self.irc_mw_gradients
        del self.irc_energies

        setattr(self, f"{prefix}_is_converged", self.converged)
        setattr(self, f"{prefix}_energy_increased", self.energy_increased)
        setattr(self, f"{prefix}_energy_converged", self.energy_converged)
        setattr(self, f"{prefix}_cycle", self.cur_cycle)
        self.dump_ends(".", prefix, getattr(self, mw_coords_name))

    def report_conv_thresholds(self):
        threshs = [
            f"\t     rms(|gradient|) <= {self.rms_grad_thresh:.6f} E_h a_0⁻¹",
            f"\t             Δenergy <= {self.energy_thresh:.6f} E_h",
        ]
        # Drop hard rms grad item
        if self.hard_rms_grad_thresh is not None:
            threshs.insert(
                1,
                f"\thard rms(|gradient|) <= {self.hard_rms_grad_thresh:.6f} E_h a_0⁻¹",
            )
        print(
            "Convergence thresholds (non mass-weighted gradient):\n"
            + "\n".join(threshs)
            + "\n"
        )

    def run(self):
        self.report_conv_thresholds()
        # Calculate data at TS and create backup
        self.ts_coords = self.coords.copy()
        self.ts_mw_coords = self.mw_coords.copy()
        print("Calculating energy and gradient at TS.")
        self.ts_gradient = self.gradient.copy()
        self.ts_mw_gradient = self.mw_gradient.copy()
        self.ts_energy = self.energy

        ts_grad_norm = np.linalg.norm(self.ts_gradient)
        ts_grad_max = np.abs(self.ts_gradient).max()
        ts_grad_rms = rms(self.ts_gradient)

        self.log(
            "Transition state (TS):\n"
            f"\t    energy={self.ts_energy:.6f} au\n"
            f"\tnorm(grad)={ts_grad_norm:.6f}\n"
            f"\t max(grad)={ts_grad_max:.6f}\n"
            f"\t rms(grad)={ts_grad_rms:.6f}"
        )

        self.init_hessian = self.geometry.hessian
        has_partial = getattr(self.geometry, "within_partial_hessian", None) is not None
        act_n_dof = (
            int(self.geometry.within_partial_hessian.get("active_n_dof", 0))
            if has_partial else 0
        )
        if has_partial:
            self._act_atoms = self.geometry.hess_active_atom_indices
            self._act_dofs = self.geometry.hess_active_dof_indices
            self.mm_inv2 = self.geometry.mm_sqrt_inv[np.ix_(self._act_dofs, self._act_dofs)]
        self.geometry.clear()
        # convert to active doFs (skip if already partial)
        if has_partial:
            if self.init_hessian.shape != (act_n_dof, act_n_dof):
                self.init_hessian = self.init_hessian[self._act_dofs][:, self._act_dofs]
        else:
            self.init_hessian = self.init_hessian[self._act_dofs][:, self._act_dofs]

        # For forward/backward runs from a TS we need an intial displacement,
        # calculated from the transition vector (imaginary mode) of the TS
        # hessian. If we need/want a Hessian for a downhill run from a
        # non-stationary point (with non-vanishing gradient) depends on the
        # actual IRC integrator (e.g. EulerPC and LQA need a Hessian).
        if not self.downhill:
            self.init_displ_plus, self.init_displ_minus = self.initial_displacement()

        print(
            "IRC length in mw. coords, max(|grad|) and rms(grad) in "
            "unweighted coordinates."
        )

        if self.forward:
            print("\n" + highlight_text("IRC - Forward") + "\n")
            self.irc("forward")
            self.set_data("forward")
            if isinstance(self.mw_hessian, torch.Tensor):
                self.forward_mw_hessian = self.mw_hessian.detach().clone()
            else:
                self.forward_mw_hessian = self.mw_hessian.copy()

        # Add TS/starting data
        self.all_energies.append(self.ts_energy)
        self.all_coords.append(self.ts_coords)
        self.all_gradients.append(self.ts_gradient)
        self.all_mw_coords.append(self.ts_mw_coords)
        self.all_mw_gradients.append(self.ts_mw_gradient)
        self.ts_index = len(self.all_energies) - 1

        if self.backward:
            print("\n" + highlight_text("IRC - Backward") + "\n")
            self.irc("backward")
            self.set_data("backward")

        if self.downhill:
            print("\n" + highlight_text("IRC - Downhill") + "\n")
            self.irc("downhill")
            self.set_data("downhill")

        self.all_mw_coords = np.array(self.all_mw_coords)
        self.all_energies = np.array(self.all_energies)
        self.postprocess()
        if not self.downhill:
            self.dump_ends(".", "finished", trj=True)

        #     # Dump the whole IRC to HDF5
        #     dump_fn = self.get_path_for_fn("finished_" + self.dump_fn)
        #     self.dump_data(dump_fn, full=True)

        # Convert to arrays and free original Python lists
        for name in "all_energies all_coords all_gradients all_mw_coords all_mw_gradients".split():
            setattr(self, name, np.array(getattr(self, name)))

        # Right now self.all_mw_coords is still in mass-weighted coordinates.
        # Convert them to un-mass-weighted coordinates.
        self.all_mw_coords_umw = self.all_mw_coords / self.m_sqrt

    def postprocess(self):
        pass

    def dump_ends(self, path, prefix, coords=None, trj=False):
        if coords is None:
            coords = self.all_mw_coords
        coords = coords.copy()
        coords /= self.m_sqrt
        coords = coords.reshape(-1, len(self.atoms), 3) * BOHR2ANG
        if trj:
            trj_string = make_trj_str(self.atoms, coords, comments=self.all_energies)
            trj_fn = self.get_path_for_fn(f"{prefix}_irc_trj.xyz")
            with open(trj_fn, "w") as handle:
                handle.write(trj_string)

        first_coords = coords[0]
        first_fn = self.get_path_for_fn(f"{prefix}_first.xyz")
        with open(first_fn, "w") as handle:
            handle.write(make_xyz_str(self.atoms, first_coords))

        last_coords = coords[-1]
        first_fn = self.get_path_for_fn(f"{prefix}_last.xyz")
        with open(first_fn, "w") as handle:
            handle.write(make_xyz_str(self.atoms, last_coords))

    def get_irc_data(self):
        data_dict = {
            "energies": np.array(self.irc_energies, dtype=float),
            "coords": np.array(self.irc_coords, dtype=float),
            "gradients": np.array(self.irc_gradients, dtype=float),
            "mw_coords": np.array(self.irc_mw_coords, dtype=float),
            "mw_gradients": np.array(self.irc_mw_gradients, dtype=float),
        }
        return data_dict

    def get_full_irc_data(self):
        data_dict = {
            "energies": np.array(self.all_energies, dtype=float),
            "coords": np.array(self.all_coords, dtype=float),
            "gradients": np.array(self.all_gradients, dtype=float),
            "mw_coords": np.array(self.all_mw_coords, dtype=float),
            "mw_gradients": np.array(self.all_mw_gradients, dtype=float),
            "ts_index": np.array(self.ts_index, dtype=int),
        }
        return data_dict

    def dump_data(self, dump_fn=None, full=False):
        get_data = self.get_full_irc_data if full else self.get_irc_data
        data_dict = get_data()

        data_dict.update(
            {
                "atoms": np.array(self.atoms, dtype="S"),
                "rms_grad_thresh": np.array(self.rms_grad_thresh),
            }
        )

        if dump_fn is None:
            dump_fn = self.get_path_for_fn(self.dump_fn)

        with h5py.File(dump_fn, "w") as handle:
            for key, val in data_dict.items():
                handle.create_dataset(name=key, dtype=val.dtype, data=val)

    def get_endpoint_and_ts_geoms(self):
        assert not self.downhill, "Downhill is not yet handled"
        first = self.all_coords[0]
        last = self.all_coords[-1]
        ts = self.ts_coords.copy()
        geoms = [Geometry(self.atoms, coords) for coords in (first, ts, last)]
        return geoms



# [1] https://doi.org/10.1002/jcc.26495
#     Habershon, 2021

import numpy as np


def get_trans_torque_forces(
    mfrag,
    a_coords3d,
    b_coords3d,
    a_mats,
    b_mats,
    m,
    frags,
    N_inv,
    weight_func=None,
    skip=True,
    kappa=1,
    do_trans=True,
):
    mcoords3d = a_coords3d[mfrag]
    gm = mcoords3d.mean(axis=0)

    if weight_func is None:

        def weight_func(m, n, a, b):
            return 1

    trans_vec = np.zeros(3)
    rot_vec = np.zeros(3)
    for n, nfrag in enumerate(frags):
        if skip and (m == n):
            continue
        amn = a_mats[(m, n)]
        bnm = b_mats[(n, m)]
        for a in amn:
            for b in bnm:
                rd = b_coords3d[b] - a_coords3d[a]
                gd = a_coords3d[a] - gm
                weight = weight_func(m, n, a, b)

                rot_vec += weight * np.cross(rd, gd)
                if do_trans:
                    trans_vec += weight * abs(rd.dot(gd)) * rd / np.linalg.norm(rd)
    trans_vec *= N_inv
    rot_vec *= N_inv
    forces = kappa * (np.cross(-rot_vec, mcoords3d - gm) + trans_vec[None, :])
    return forces


class TransTorque:
    def __init__(
        self,
        frags,
        iter_frags,
        a_mats,
        b_mats,
        weight_func=None,
        skip=True,
        kappa=1.0,
        b_coords3d=None,
        do_trans=True,
    ):
        """Translational and torque forces.
        See A.4. [1], Eqs. (A3) - (A5).
        """
        self.frags = frags
        self.iter_frags = iter_frags
        self.a_mats = a_mats
        self.b_mats = b_mats
        self.weight_func = weight_func
        self.kappa = kappa
        self.skip = skip
        self.b_coords3d = b_coords3d
        self.do_trans = do_trans

        self.set_N_invs()

    def set_N_invs(self):
        Ns = np.zeros(len(self.frags))
        for m, mfrag in enumerate(self.frags):
            for n, _ in enumerate(self.iter_frags):
                if self.skip and (m == n):
                    continue
                amn = self.a_mats[(m, n)]
                bnm = self.b_mats[(n, m)]
                Ns[m] += len(amn) * len(bnm)
            Ns[m] *= 3 * len(mfrag)
        self.N_invs = np.divide(1, Ns, out=np.zeros_like(Ns), where=Ns != 0)

    def get_forces(self, atoms, coords, kappa=None):
        if kappa is None:
            kappa = self.kappa
        c3d = coords.reshape(-1, 3)
        forces = np.zeros_like(c3d)

        b_coords3d = c3d if self.b_coords3d is None else self.b_coords3d

        for m, mfrag in enumerate(self.frags):
            N_inv = self.N_invs[m]
            tt_forces = get_trans_torque_forces(
                mfrag,
                c3d,
                b_coords3d,
                self.a_mats,
                self.b_mats,
                m,
                self.iter_frags,
                N_inv,
                weight_func=self.weight_func,
                skip=self.skip,
                do_trans=self.do_trans,
                kappa=kappa,
            )
            forces[mfrag] = tt_forces

        return {"energy": 1, "forces": forces.flatten()}

    def get_forces_naive(self, atoms, coords, kappa=None):
        if kappa is None:
            kappa = self.kappa

        AR = self.a_mats
        Ns = np.zeros(len(self.frags))
        for m, mfrag in enumerate(self.frags):
            for n, nfrag in enumerate(self.frags):
                if m == n:
                    continue
                Ns[m] += len(AR[(n, m)]) * len(AR[(m, n)])
            Ns[m] *= 3 * len(mfrag)

        c3d = coords.reshape(-1, 3).copy()
        f3d = np.zeros_like(c3d)
        vts = np.zeros((len(self.frags), 3))
        vrs = np.zeros((len(self.frags), 3))
        for m, mfrag in enumerate(self.frags):
            gm = c3d[mfrag].mean(axis=0)
            c3dm = c3d[mfrag]
            for n, nfrag in enumerate(self.frags):
                if m == n:
                    continue
                for a in AR[(m, n)]:
                    for b in AR[(n, m)]:
                        rdiff = c3d[b] - c3d[a]
                        rdiffn = np.linalg.norm(rdiff)
                        rg = c3d[a] - gm
                        quot = rdiff / rdiffn
                        dot = rdiff.dot(rg)
                        vts[m] += abs(dot) * quot
                        vrs[m] += np.cross(rdiff, rg)
            vts[m] /= Ns[m]
            vrs[m] /= Ns[m]
            if not self.do_trans:
                vts[m] = 0.0
            f3d[mfrag] = kappa * (-np.cross(vrs[m], c3dm-gm[None, :]) + vts[m])

        forces = f3d.flatten()
        return {"energy": 1, "forces": forces}

    # def get_forces(self, atoms, coords, kappa=None):
        # return self.get_forces_naive(atoms, coords, kappa=kappa)



# pdb2reaction/add_elem_info.py

"""
Repair PDB element symbols (columns 77-78) with Biopython inference.

Example:
    pdb2reaction add-elem-info -i 1abc.pdb

For detailed documentation, see: docs/add_elem_info.md
"""

from __future__ import annotations

import argparse
import collections
import os
import re
import sys
from pathlib import Path
from typing import Optional

import click
from Bio.PDB import PDBParser, PDBIO

# Reuse residue/ion dictionaries from extract.py to keep definitions in sync
from .extract import AMINO_ACIDS, ION, WATER_RES

# -----------------------------
# Element symbols (IUPAC, 1–118)
# -----------------------------
ELEMENTS: set[str] = {
    "H","He","Li","Be","B","C","N","O","F","Ne","Na","Mg","Al","Si","P","S","Cl","Ar",
    "K","Ca","Sc","Ti","V","Cr","Mn","Fe","Co","Ni","Cu","Zn","Ga","Ge","As","Se","Br","Kr",
    "Rb","Sr","Y","Zr","Nb","Mo","Tc","Ru","Rh","Pd","Ag","Cd","In","Sn","Sb","Te","I","Xe",
    "Cs","Ba","La","Ce","Pr","Nd","Pm","Sm","Eu","Gd","Tb","Dy","Ho","Er","Tm","Yb","Lu",
    "Hf","Ta","W","Re","Os","Ir","Pt","Au","Hg","Tl","Pb","Bi","Po","At","Rn","Fr","Ra",
    "Ac","Th","Pa","U","Np","Pu","Am","Cm","Bk","Cf","Es","Fm","Md","No","Lr","Rf","Db",
    "Sg","Bh","Hs","Mt","Ds","Rg","Cn","Fl","Lv","Ts","Og"
}

# Common residue classes
PROTEIN_RES = set(AMINO_ACIDS.keys())
NUCLEIC_RES = {
    "DA","DT","DG","DC","DI",
    "A","U","G","C","I",
}

# -----------------------------
# Helper: normalize strings to element symbols
# -----------------------------
_re_letters = re.compile(r"[A-Za-z]+")

def _normalize_symbol(s: str) -> Optional[str]:
    """Remove non-letters; prefer a 2-letter match, then 1-letter, against known elements.
    Returns the correctly cased symbol if matched.
    Treat deuterium 'D' as hydrogen 'H' (PDB often uses D interchangeably with H).
    """
    if not s:
        return None
    m = _re_letters.findall(s)
    if not m:
        return None
    letters = "".join(m)
    if len(letters) >= 2:
        cand2 = (letters[:2][0].upper() + letters[:2][1].lower())
        if cand2 in ELEMENTS:
            return cand2
    cand1 = letters[0].upper()
    if cand1 in ELEMENTS:
        return cand1
    if letters[0].upper() == "D":
        return "H"
    return None

def _symbol_from_resname(resname: str) -> Optional[str]:
    """Extract an element symbol from an ion residue name (e.g., CA, FE2, Cl-, YB2)."""
    res = resname.strip()
    return _normalize_symbol(res)

def _default_out_pdb_path(in_pdb: str) -> str:
    """Default output path when -o/--out is omitted:
    replace trailing '.pdb' (case-insensitive) with '_add_elem.pdb';
    if no trailing '.pdb', append '_add_elem.pdb'.
    """
    p = Path(in_pdb)
    name = p.name
    if name.lower().endswith(".pdb"):
        name = name[:-4] + "_add_elem.pdb"
    else:
        name = name + "_add_elem.pdb"
    return str(p.with_name(name))

# -----------------------------
# Element inference (use residue to disambiguate)
# -----------------------------
def guess_element(atom_name: str, resname: str, is_het: bool) -> Optional[str]:
    """
    Infer the element from atom name + residue name.
    Priority:
      1) Ion residues: prefer the residue name (NH4 / H3O+ handled per-atom as H/N/O)
      2) Polymers (protein/nucleic acid) and water: follow convention (H/C/N/O/S/P/Se)
      3) Other ligands: use atom-name prefix; then normalization
      4) Unresolved → None
    """
    name_u = atom_name.strip().upper()
    res_u = resname.strip().upper()

    # 1) Ion residues — strongly prefer the residue-derived element
    if res_u in {k.upper() for k in ION.keys()}:
        # Polyatomic ions (NH4, H3O+, …): decide per atom name (treat D* as H)
        if name_u.startswith(("H", "D")):
            return "H"
        if name_u.startswith("N"):
            return "N"
        if name_u.startswith("O"):
            return "O"
        # Monatomic metals/halogens: from residue name
        sym = _symbol_from_resname(res_u)
        if sym:
            return sym

    # 2) Polymers (protein/nucleic) / water
    is_protein = res_u in PROTEIN_RES
    is_nucl = res_u in NUCLEIC_RES
    is_water = res_u in WATER_RES
    if is_protein or is_nucl or is_water:
        # Water: only O and H (treat D* as H)
        if is_water:
            if name_u.startswith(("H", "D")):
                return "H"
            return "O"

        # Hydrogen (including D*)
        if name_u.startswith(("H", "D")):
            return "H"

        # Selenium (e.g., selenomethionine/selenocysteine)
        if name_u.startswith("SE"):
            return "Se"

        # P, N, O, S map directly by first letter
        if name_u.startswith("P"):
            return "P"
        if name_u.startswith("N"):
            return "N"
        if name_u.startswith("O"):
            return "O"
        if name_u.startswith("S"):
            return "S"

        # Carbon for Cα/sidechain labels (CA, CB, CG, CD, CE, CZ, CH*, etc.)
        if name_u.startswith("C"):
            return "C"

        # Fallback to normalization (rare halogens or atypical labels)
        sym = _normalize_symbol(name_u)
        if sym:
            return sym

    # 3) Non-polymers (ligands / cofactors)
    # H, C, N, O, P map directly by first letter (similar to polymers)
    if name_u.startswith(("H", "D")):
        return "H"
    if name_u.startswith("C") and not name_u.startswith("CL"):
        return "C"
    if name_u.startswith("N"):
        return "N"
    if name_u.startswith("O"):
        return "O"
    if name_u.startswith("P"):
        return "P"

    sym = _normalize_symbol(name_u)
    if sym:
        return sym

    # 4) Unresolved
    return None

def _get_atom_serial(atom) -> Optional[int]:
    """Safely obtain the serial number from a Biopython Atom, handling version differences."""
    sn = getattr(atom, "serial_number", None)
    if sn is None and hasattr(atom, "get_serial_number"):
        try:
            sn = atom.get_serial_number()
        except Exception:
            sn = None
    return sn

# -----------------------------
# Main processing
# -----------------------------
def assign_elements(in_pdb: str, out_pdb: Optional[str], overwrite: bool = False) -> None:
    # If an explicit output path is provided, never overwrite in-place even when --overwrite is
    # passed. This keeps -o/-\-out as the higher-priority choice.
    effective_overwrite = overwrite and out_pdb is None

    parser = PDBParser(QUIET=True)
    structure_id = os.path.splitext(os.path.basename(in_pdb))[0]
    structure = parser.get_structure(structure_id, in_pdb)

    total = 0
    assigned_or_updated = 0   # atoms whose inferred element differs from the parsed value
    unknown = []              # could not infer (left unchanged)

    by_element = collections.Counter()

    for model in structure:
        for chain in model:
            for residue in chain:
                hetflag = residue.id[0].strip()  # '' (standard); 'W' = water; 'H_' = HETATM
                is_het = (hetflag != "")
                resname = residue.get_resname()
                for atom in residue:
                    total += 1
                    name = atom.get_name()

                    sym = guess_element(name, resname, is_het)
                    if sym is None:
                        serial = _get_atom_serial(atom)
                        unknown.append((model.id, chain.id, residue.id, resname, name, serial))
                        continue

                    prev = getattr(atom, "element", None)
                    atom.element = sym
                    by_element[sym] += 1
                    if prev != sym:
                        assigned_or_updated += 1

    io = PDBIO()
    io.set_structure(structure)
    if effective_overwrite:
        out_path = in_pdb
    else:
        out_path = out_pdb if out_pdb else _default_out_pdb_path(in_pdb)
    io.save(out_path)

    # Summary
    click.echo(f"[OK] Wrote: {out_path}")
    click.echo(f"  total atoms                 : {total}")
    click.echo(f"  assigned/updated            : {assigned_or_updated}")
    if by_element:
        top = ", ".join(f"{k}:{v}" for k, v in by_element.most_common())
        click.echo(f"  assignment breakdown        : {top}")
    if unknown:
        click.echo(f"[WARN] Could not confidently assign {len(unknown)} atoms; left unchanged.")
        for (mid, chid, resid, resn, aname, serial) in unknown[:50]:
            if isinstance(resid, tuple):
                resseq = resid[1]
                icode = resid[2].strip()
            else:
                resseq, icode = "?", ""
            s_str = f" serial {serial}" if serial is not None else ""
            click.echo(f"    model {mid} chain {chid} {resn} {resseq}{icode} : {aname}{s_str}")
    if len(unknown) > 50:
        click.echo("    ... (truncated) ...")


def main():
    ap = argparse.ArgumentParser(
        description="Add/repair element columns (77–78) in a PDB using Biopython."
    )
    ap.add_argument(
        "-i",
        "--input",
        dest="in_pdb",
        required=True,
        help="Input PDB filepath.",
    )
    ap.add_argument(
        "-o",
        "--out",
        help='Output PDB filepath (default: replace ".pdb" with "_add_elem.pdb"; when provided, --overwrite is ignored).',
    )
    ap.add_argument(
        "--overwrite",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="Overwrite the input file in-place when -o/--out is omitted.",
    )
    args = ap.parse_args()

    if not os.path.isfile(args.in_pdb):
        click.echo(f"[ERR] Input not found: {args.in_pdb}", err=True)
        sys.exit(1)

    try:
        assign_elements(args.in_pdb, args.out, overwrite=args.overwrite)
    except Exception as e:
        click.echo(f"[ERR] Failed: {e}", err=True)
        sys.exit(2)

# -----------------------------
# Click subcommand (pdb2reaction add-elem-info)
# -----------------------------
@click.command(
    help="Add/repair element columns (77–78) in a PDB using Biopython.",
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.option(
    "-i", "--input",
    "in_pdb",
    type=click.Path(path_type=Path, exists=True, dir_okay=False),
    required=True,
    help="Input PDB filepath.",
)
@click.option(
    "-o", "--out",
    "out_pdb",
    type=click.Path(path_type=Path, dir_okay=False),
    default=None,
    help='Output PDB filepath (default: replace ".pdb" with "_add_elem.pdb"; when provided, --overwrite is ignored).',
)
@click.option(
    "--overwrite/--no-overwrite",
    default=False,
    show_default=True,
    help="Overwrite the input file in-place when -o/--out is omitted.",
)
def cli(in_pdb: Path, out_pdb: Optional[Path], overwrite: bool) -> None:
    """Click wrapper to run via the `pdb2reaction add-elem-info` subcommand."""
    try:
        assign_elements(str(in_pdb), (str(out_pdb) if out_pdb else None), overwrite=overwrite)
    except SystemExit as e:
        raise e
    except Exception as e:
        click.echo(f"[ERR] Failed: {e}", err=True)
        sys.exit(2)



import logging

__all__ = [
    "AdaptiveNEB",
    "ChainOfStates",
    "FreeEndNEB",
    "FreezingString",
    "GrowingNT",
    "GrowingString",
    "NEB",
    "SimpleZTS",
]

logger = logging.getLogger("cos")
logger.setLevel(logging.DEBUG)
# delay = True prevents creation of empty logfiles
handler = logging.FileHandler("cos.log", mode="w", delay=True)
# fmt_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
fmt_str = "%(message)s"
formatter = logging.Formatter(fmt_str)
handler.setFormatter(formatter)
logger.addHandler(handler)



import numpy as np

from pysisyphus.irc.IRC import IRC


# [1] https://aip.scitation.org/doi/pdf/10.1063/1.462674?class=pdf
# [2] https://www.sciencedirect.com/science/article/pii/S0009261406015673


class RK4(IRC):

    def get_k(self, mw_coords):
        self.mw_coords = mw_coords
        grad = self.mw_gradient
        direction = -grad / np.linalg.norm(grad)
        k = self.step_length * direction
        return k

    def step(self):
        # Eq. (25 - 29) in [1]
        mw_coords_1 = self.mw_coords.copy()
        k1 = self.get_k(mw_coords_1)

        mw_coords_2 = mw_coords_1 + 0.5*k1
        k2 = self.get_k(mw_coords_2)

        mw_coords_3 = mw_coords_1 + 0.5*k2
        k3 = self.get_k(mw_coords_3)

        mw_coords_4 = mw_coords_1 + k3
        k4 = self.get_k(mw_coords_4)

        step = (k1 + 2*k2 + 2*k3 + k4) / 6
        step_norm = np.linalg.norm(step)
        self.log(f"norm(step)={step_norm:.6f}")
        self.mw_coords = mw_coords_1 + step



import re

import numpy as np

from pysisyphus.constants import BOHR2ANG


def make_xyz_str(atoms, coords, comment=""):
    assert len(atoms) == len(coords)
    atoms = [a.capitalize() for a in atoms]

    coord_fmt = "{: 03.8f}"
    line_fmt = "{:>3s} " + " ".join(
        [
            coord_fmt,
        ]
        * 3
    )

    body = [line_fmt.format(a, *xyz) for a, xyz in zip(atoms, coords)]
    body = "\n".join(body)

    return f"{len(atoms)}\n{comment}\n{body}"


def make_trj_str(atoms, coords_list, comments=None):
    if comments is None:
        comments = ["" for _ in coords_list]
    xyz_strings = [
        make_xyz_str(atoms, coords, comment)
        for coords, comment in zip(coords_list, comments)
    ]
    return "\n".join(xyz_strings)


def coords_to_trj(trj_fn, atoms, coords_list, comments=None):
    coords = np.array(coords_list)
    coords = coords.reshape(-1, len(atoms), 3) * BOHR2ANG
    trj_str = make_trj_str(atoms, coords, comments)
    with open(trj_fn, "w") as handle:
        handle.write(trj_str)
    return trj_fn


def make_trj_str_from_geoms(geoms, comments=None, energy_comments=False):
    atoms = geoms[0].atoms
    coords_list = [geom.coords3d * BOHR2ANG for geom in geoms]
    if energy_comments and comments is None:
        comments = [str(geom._energy) for geom in geoms]
    elif comments is not None:
        assert len(comments) == len(geoms)

    return make_trj_str(atoms, coords_list, comments)


def write_geoms_to_trj(geoms, fn, comments=None):
    trj_str = make_trj_str_from_geoms(geoms, comments)
    with open(fn, "w") as handle:
        handle.write(trj_str)


def split_xyz_str(xyz_str):
    """Example:

    xyz:
     1

     X -1.2 1.4 0.0
     1

     X 2.0 4.0 0.0

    """
    float_ = r"([\+\d\-\.]+)"
    header_re = re.compile(r"(\d+)")
    coord_re = re.compile(fr"[a-zA-Z]+\s+{float_}\s+{float_}\s+{float_}")

    lines = [l.strip() for l in xyz_str.strip().split("\n")]

    lines_remaining = len(lines)
    cur_line = 0
    valid_xyz_strs = list()
    while lines_remaining:
        header_mobj = header_re.match(lines[cur_line])
        expect_lines = int(header_mobj[1])
        slice_ = slice(
            cur_line + 2, cur_line + 2 + expect_lines
        )  # lgtm [py/hash-unhashable-value]
        check_lines = lines[slice_]
        assert len(check_lines) == expect_lines
        assert all([coord_re.match(line.strip()) for line in check_lines])

        valid_xyz_strs.append(str(expect_lines) + "\n\n" + "\n".join(check_lines))

        lines_read = expect_lines + 2
        lines_remaining -= lines_read
        cur_line += lines_read
    atoms_coords = [parse_xyz_str(_, with_comment=False) for _ in valid_xyz_strs]
    return atoms_coords


def parse_xyz_str(xyz_str, with_comment):
    """Parse a xyz string.

    Paramters
    ---------
    xyz_str : str
        The contents of a .xyz file.
    with_comment : bool
        Return comment line if True.

    Returns
    -------
    atoms : list
        List of length N (N = number of atoms) holding the
        element symbols.
    coords: np.array
        An array of shape (N, 3) holding the xyz coordinates.
    comment_line : str, optional
        Comment line if with_comment argument was True.
    """

    xyz_lines = xyz_str.strip().split("\n")
    comment_line = xyz_lines[1]

    # Only consider the first four items on a line
    atom_num = int(xyz_lines[0])
    atoms_present = len(xyz_lines) - 2
    assert len(xyz_lines) == atom_num + 2, \
        f"Expected {atom_num} atoms, but found only {atoms_present}!"
    atoms_coords = [
        line.strip().split()[:4] for line in xyz_str.strip().split("\n")[2:]
    ]
    atoms, coords = zip(*[(a, c) for a, *c in atoms_coords])
    coords = np.array(coords, dtype=float)
    if with_comment:
        return atoms, coords, comment_line
    else:
        return atoms, coords


def parse_xyz_file(xyz_fn, with_comment=False):
    with open(xyz_fn) as handle:
        xyz_str = handle.read()

    return parse_xyz_str(xyz_str, with_comment)


def parse_trj_file(trj_fn, with_comments=False):
    with open(trj_fn) as handle:
        trj_str = handle.read()

    return parse_trj_str(trj_str, with_comments)


def parse_trj_str(trj_str, with_comments=False):
    trj_lines = trj_str.strip().split("\n")
    number_of_atoms = int(trj_lines[0].strip())
    xyz_lines = number_of_atoms + 2
    # Split the trj file in evenly sized strings
    xyz_strs = [
        "\n".join(trj_lines[i : i + xyz_lines])
        for i in range(0, len(trj_lines), xyz_lines)
    ]
    xyzs = [parse_xyz_str(xyz_str, with_comments) for xyz_str in xyz_strs]

    assert len(xyzs) == (len(trj_lines) / xyz_lines)
    return xyzs



import numpy as np

from pysisyphus.constants import BOHR2M, AU2J, AMU2KG
from pysisyphus.irc.IRC import IRC
from pysisyphus.TableFormatter import TableFormatter


# [1] https://pubs.acs.org/doi/10.1021/jp012125b
#     Following Reaction Pathways Using a Damped Classical Trajectory Algorithm
#     Hratchian, Schlegel, 2002


class DampedVelocityVerlet(IRC):

    def __init__(self, geometry, v0=0.04, dt0=0.5, error_tol=0.003,
                 max_cycles=150, **kwargs):
        super().__init__(geometry, max_cycles=max_cycles, **kwargs)

        self.v0 = v0
        self.error_tol = error_tol
        self.dt0 = dt0

        step_header = "damping dt dt_new error".split()
        step_fmts = [".2f", ".4f", ".4f", ".3E"]
        self.step_formatter = TableFormatter(step_header, step_fmts, 10)

    def mw_grad_to_acc(self, mw_grad):
        """Takes care of the units for the mass-weighted gradient.
        Converts units of a mass-weighted gradient [Hartree/(Bohr*amu)]
        to units of acceleration [sqrt(amu)*Bohr/fs²]. The 1e30 comes
        from converting second² to femto second²."""
        return mw_grad * AU2J / AMU2KG / BOHR2M**2 / 1e30

    def prepare(self, direction):
        # In contrast to the paper we don't start from the TS geometry but
        # instead do an initial displacement along the imaginary mode as for
        # the other IRC classes. So there should always be a non-vanishing
        # gradient after the prepare call.
        super().prepare(direction)

        acceleration = -self.mw_grad_to_acc(self.mw_gradient)
        # init_velocity = acceleration
        init_velocity, _ = self.damp_velocity(acceleration)

        self.velocities = [init_velocity]
        self.accelerations = [acceleration]
        self.time_steps = [self.dt0]

    def damp_velocity(self, velocity):
        # Eq. (4) in [1]
        damping_factor = self.v0 / np.linalg.norm(velocity)
        self.log(f"Damping factor={damping_factor:.6f}")
        damped_velocity = velocity * damping_factor
        return damped_velocity, damping_factor

    def estimate_error(self, new_mw_coords):
        self.log("Error estimation")
        # See Fig. 1 and Eq. (5) in [1]
        cur_time_step = self.time_steps[-1]
        prev_time_step = self.time_steps[-2]
        time_step_sum = prev_time_step + cur_time_step
        self.log(f"\tSum of cur. and prev. timestep: {time_step_sum:.6f} fs")
        # Calculation of x' coordinates
        # irc_mw_coords
        #   -1: current coords  <- x_i-1 in the paper
        #   -2: previous coords <- x_i-2 in the paper
        # velocities/accelerations
        #   -1: new velo./acc. for next time step
        #   -2: velo./acc. that yielded the proposed coords x_i
        #   -3: velo./acc. that yielded previous structure <- v_i-2 and a_i-2 in the paper
        ref_coords = (
            self.irc_mw_coords[-2]
            # TODO: This should probably be -3 but it seems to give inferior results
            + self.velocities[-2]*time_step_sum
            + 0.5*self.accelerations[-2]*time_step_sum**2
        )
        diff = new_mw_coords - ref_coords
        diff /= np.sqrt(self.geometry.masses_rep)
        largest_component = np.max(np.abs(diff))
        norm = np.linalg.norm(diff)
        # Take either the largest component of the difference vector
        # or the norm of the difference vector.
        self.log("\tdiff=Proposed coords - Reference coords")
        self.log(f"\tmax(|diff|)={largest_component:.6f}")
        self.log(f"\ttnorm(diff)={norm:.6f}")
        # estimated_error = max(largest_component, norm)
        estimated_error = max(largest_component, norm)
        self.log(f"\testimated error={estimated_error:.6f}")
        return estimated_error

    def step(self):
        prev_acceleration = self.accelerations[-1]
        prev_velocity = self.velocities[-1]
        time_step = self.time_steps[-1]

        # Get new acceleration
        acceleration = -self.mw_grad_to_acc(self.mw_gradient)

        acc_normed = acceleration/np.linalg.norm(acceleration)
        prev_vel_normed = prev_velocity / np.linalg.norm(prev_velocity)
        ovlp = acc_normed @ prev_vel_normed
        self.log(f"a @ v_i-1={ovlp:.8f}")
        self.accelerations.append(acceleration)

        # Calculate new coords and velocity
        # Eq. (2) in [1]
        new_mw_coords = (self.mw_coords
                         + prev_velocity*time_step
                         + 0.5*prev_acceleration*time_step**2
        )

        # Update velocity
        velocity = prev_velocity + 0.5*(prev_acceleration+acceleration)*time_step
        # Damp velocity
        damped_velocity, damping_factor = self.damp_velocity(velocity)
        self.velocities.append(damped_velocity)

        if self.cur_cycle == 0:
            # No error estimation in the first step
            estimated_error = self.error_tol
        else:
            estimated_error = self.estimate_error(new_mw_coords)

        # Get next time step from error estimation
        new_time_step = time_step * (self.error_tol / estimated_error)**(1/3)
        # Constrain time step between 0.0025 fs and 3.0 fs
        new_time_step = min(new_time_step, 3.)
        new_time_step = max(new_time_step, 0.025)
        self.time_steps.append(new_time_step)
        self.log(f"\tCur. time step={time_step:.6f} fs")
        self.log(f"\tNext time step={new_time_step:.6f} fs")

        # Set new coordinates
        self.mw_coords = new_mw_coords



import abc
import logging
from math import exp

import numpy as np

from pysisyphus.elem_data import COVALENT_RADII as CR
from pysisyphus.helpers_pure import hash_arr
from pysisyphus.linalg import norm3


class Primitive(metaclass=abc.ABCMeta):
    def __init__(self, indices, periodic=False, calc_kwargs=None, cache=False):
        self.indices = list(indices)
        self.periodic = periodic
        if calc_kwargs is None:
            calc_kwargs = ()
        self.calc_kwargs = calc_kwargs
        self.cache = cache

        self.logger = logging.getLogger("internal_coords")
        self.val_cache = {}
        self.grad_cache = {}

    def log(self, msg, lvl=logging.DEBUG):
        self.logger.log(lvl, msg)

    def log_dbg(self, msg):
        self.log(msg, lvl=logging.DEBUG)

    @staticmethod
    def parallel(u, v, thresh=1e-6):
        dot = u.dot(v) / (norm3(u) * norm3(v))
        return (1 - abs(dot)) < thresh

    @staticmethod
    def _get_cross_vec(coords3d, indices):
        m, o, n = indices
        # Select initial vector for cross product, similar to
        # geomeTRIC. It must NOT be parallel to u and/or v.
        x_dash = coords3d[n] - coords3d[m]
        x = x_dash / norm3(x_dash)
        cross_vecs = np.eye(3)
        min_ind = np.argmin([np.dot(cv, x) ** 2 for cv in cross_vecs])
        return cross_vecs[min_ind]

    def set_cross_vec(self, coords3d, indices):
        self.cross_vec = self._get_cross_vec(coords3d, self.indices)
        self.log(f"Cross vector for {self} set to {self.cross_vec}")

    @abc.abstractmethod
    def _calculate(*, coords3d, indices, gradient, **kwargs):
        pass

    @abc.abstractmethod
    def _weight(self, atoms, coords3d, indices, f_damping):
        pass

    def weight(self, atoms, coords3d, f_damping=0.12):
        return self._weight(atoms, coords3d, self.indices, f_damping)

    @staticmethod
    def rho(atoms, coords3d, indices):
        i, j = indices
        distance = norm3(coords3d[i] - coords3d[j])
        cov_rad_sum = CR[atoms[i].lower()] + CR[atoms[j].lower()]
        return exp(-(distance / cov_rad_sum - 1))

    # def calculate(self, coords3d, indices=None, gradient=False):
        # if indices is None:
            # indices = self.indices

        # # Gather calc_kwargs
        # calc_kwargs = {key: getattr(self, key) for key in self.calc_kwargs}

        # return self._calculate(
            # coords3d=coords3d,
            # indices=indices,
            # gradient=gradient,
            # **calc_kwargs,
        # )

    def calculate(self, coords3d, indices=None, gradient=False):
        if indices is None:
            indices = self.indices

        if self.cache:
            cur_hash = hash_arr(coords3d[indices], precision=8)
            try:
                val = self.val_cache[cur_hash]
                if gradient:
                    grad = self.grad_cache[cur_hash]
                    self.log_dbg(f"Returning cached value & gradient for hash '{cur_hash}'.")
                    return val, grad
                else:
                    self.log_dbg(f"Returning cached value for hash '{cur_hash}'.")
                    return val
            except KeyError:
                self.log_dbg(f"Hash '{cur_hash}' is not yet cached.")

        # Gather calc_kwargs
        calc_kwargs = {key: getattr(self, key) for key in self.calc_kwargs}

        results =  self._calculate(
            coords3d=coords3d,
            indices=indices,
            gradient=gradient,
            **calc_kwargs,
        )

        if self.cache:
            if gradient:
                val, grad = results
                self.val_cache[cur_hash] = val
                self.grad_cache[cur_hash] = grad
            else:
                self.val_cache[cur_hash] = results
        return results

    def jacobian(self, coords3d, indices=None):
        if indices is None:
            indices = self.indices

        # Gather calc_kwargs
        calc_kwargs = {key: getattr(self, key) for key in self.calc_kwargs}

        return self._jacobian(
            coords3d=coords3d,
            indices=indices,
            **calc_kwargs,
        )

    def __str__(self):
        return f"{self.__class__.__name__}({self.indices})"

    def __repr__(self):
        return self.__str__()



# pdb2reaction/align_freeze_atoms.py

"""
Rigid alignment and staged scan+relaxation utilities for pysisyphus Geometry objects.

Main Functions:
    - kabsch_R_t: Row-vector Kabsch solver for (N,3) point sets.
    - align_second_to_first_kabsch_inplace: Rigidly align g_mob to g_ref in place.
    - scan_freeze_atoms_toward_target_inplace: Move freeze_atoms toward reference with relaxation.
    - align_and_refine_pair_inplace: Rigid alignment + scan/relaxation for a geometry pair.
    - align_and_refine_sequence_inplace: Apply pair procedure along a sequence of geometries.

Units: User-facing distances in Å; internal coordinates in bohr. Indices are 0-based.
"""

from __future__ import annotations

from pathlib import Path
from typing import List, Optional, Sequence, Tuple, Dict, Any

import numpy as np
import click

# pysisyphus
from pysisyphus.optimizers.LBFGS import LBFGS
from pysisyphus.optimizers.exceptions import OptimizationError, ZeroStepLength
from pysisyphus.constants import BOHR2ANG

# Support both relative and absolute import paths for the backend factory
try:
    # Within the package
    from .backends import create_calculator
except Exception:
    # Direct execution, etc.
    from pdb2reaction.backends import create_calculator

from .utils import as_list


# =============================================================================
# Math utilities (row-vector convention: Q @ R + t)
# =============================================================================

def kabsch_R_t(P: np.ndarray, Q: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Row-vector Kabsch: find R, t minimizing ||(Q @ R + t) - P|| for (N, 3).
    """
    P = np.asarray(P, float)
    Q = np.asarray(Q, float)
    if P.shape != Q.shape or P.ndim != 2 or P.shape[1] != 3:
        raise ValueError("Kabsch expects P, Q with shape (N, 3).")
    mu_P, mu_Q = P.mean(0), Q.mean(0)
    Pc, Qc = P - mu_P, Q - mu_Q
    H = Pc.T @ Qc
    U, _, Vt = np.linalg.svd(H)
    R = Vt.T @ U.T
    if np.linalg.det(R) < 0.0:
        Vt[-1] *= -1.0
        R = Vt.T @ U.T
    t = mu_P - mu_Q @ R
    return R, t


def _rodrigues(axis_unit: np.ndarray, theta: float) -> np.ndarray:
    """
    3×3 rotation matrix for a rotation of angle `theta` about `axis_unit`.
    """
    u = np.asarray(axis_unit, float)
    n = np.linalg.norm(u)
    if n < 1e-16:
        return np.eye(3)
    u /= n
    ux, uy, uz = u
    K = np.array([[0.0, -uz,  uy],
                  [uz,  0.0, -ux],
                  [-uy, ux,  0.0]], float)
    I = np.eye(3)
    return I + np.sin(theta) * K + (1.0 - np.cos(theta)) * (K @ K)


def _rotation_align_vectors(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """
    3×3 rotation (column-vector convention) that maps vector `a` to `b`.
    When applying to row-vectors, use the transpose of the returned matrix.
    """
    a = np.asarray(a, float)
    b = np.asarray(b, float)
    an = np.linalg.norm(a)
    bn = np.linalg.norm(b)
    if an < 1e-16 or bn < 1e-16:
        return np.eye(3)
    a /= an
    b /= bn
    v = np.cross(a, b)
    c = float(np.clip(a.dot(b), -1.0, 1.0))
    s = np.linalg.norm(v)
    if s < 1e-12:
        if c > 0.0:
            return np.eye(3)
        # 180°: choose any axis orthogonal to `a`
        tmp = np.array([1.0, 0.0, 0.0]) if abs(a[0]) <= 0.9 else np.array([0.0, 1.0, 0.0])
        axis = np.cross(a, tmp)
        axis /= (np.linalg.norm(axis) + 1e-16)
        return _rodrigues(axis, np.pi)
    axis = v / s
    theta = np.arctan2(s, c)
    return _rodrigues(axis, theta)


def _orth_proj_perp(u: np.ndarray) -> np.ndarray:
    """
    3×3 projector onto the plane perpendicular to vector `u`.
    """
    u = np.asarray(u, float)
    n = np.linalg.norm(u)
    if n < 1e-16:
        return np.eye(3)
    u = u / n
    return np.eye(3) - np.outer(u, u)


# =============================================================================
# Geometry utilities
# =============================================================================

def _coords3d(geom) -> np.ndarray:
    """
    Return (N, 3) coordinates in bohr (float).
    """
    return np.array(geom.coords3d, float)


def _rmsd(A: np.ndarray, B: np.ndarray) -> float:
    """
    Compute RMSD in Å (inputs are in bohr; conversion is applied internally).
    """
    A = np.asarray(A, float) * BOHR2ANG
    B = np.asarray(B, float) * BOHR2ANG
    return float(np.sqrt(np.mean(np.sum((A - B) ** 2, axis=1)))) if len(A) else float("nan")


def _set_all_coords_disabling_freeze(geom, coords3d_bohr: np.ndarray) -> None:
    """
    Temporarily clear `freeze_atoms`, set all coordinates (bohr), then restore.
    """
    old = np.array(getattr(geom, "freeze_atoms", []), int)
    try:
        geom.freeze_atoms = np.array([], int)
        geom.set_coords(np.asarray(coords3d_bohr, float).reshape(-1), cartesian=True)
    finally:
        geom.freeze_atoms = old


def _attach_calc_if_needed(geom, shared_calc=None, *, charge=0, spin=1,
                           model="uma-s-1p1", device="auto") -> None:
    """
    Set `shared_calc` when provided; otherwise attach UMA if no calculator is present.
    """
    try:
        has_calc = getattr(geom, "calculator", None) is not None
    except Exception:
        has_calc = False
    if shared_calc is not None:
        geom.set_calculator(shared_calc)
    elif not has_calc:
        geom.set_calculator(create_calculator(charge=charge, spin=spin, model=model, device=device))


def _freeze_union(g_ref, g_mob, n_atoms: Optional[int] = None) -> List[int]:
    """
    Union of `freeze_atoms` from `g_ref` and `g_mob` (0-based).
    If `n_atoms` is given, out-of-range indices are removed. Returns [] if empty.
    """
    fa0 = getattr(g_ref, "freeze_atoms", None)
    fa1 = getattr(g_mob, "freeze_atoms", None)

    cand = sorted(set(int(i) for i in (as_list(fa0) + as_list(fa1))))
    if n_atoms is None:
        return cand
    good = [i for i in cand if 0 <= i < int(n_atoms)]
    return good


# =============================================================================
# Rigid alignment: special handling for union length = 1/2 → Kabsch otherwise
# =============================================================================

def align_second_to_first_kabsch_inplace(g_ref, g_mob,
                                         *, verbose: bool = True) -> Dict[str, Any]:
    """
    Rigidly align `g_mob` to `g_ref` (update coordinates of `g_mob` in place).

    Returns a dict with keys: {before_A, after_A, n_used, mode}
      - mode: "one_anchor" | "two_anchor" | "kabsch"

    Behavior (selection = union of freeze atoms across the pair):
      * length = 1 atom: minimize the all-atom RMSD using rotations about the
        single anchor point only. RMSD is reported on all atoms.
      * length = 2 atoms: align the axis defined by the two anchors and optimize
        the rotation around that axis. RMSD is reported on all atoms.
        If the axis is degenerate, falls back to standard Kabsch.
      * otherwise: solve Kabsch with the union selection (or all atoms if empty)
        and apply the resulting transform to all atoms. RMSD is reported on the
        same selection used by Kabsch.
    """
    P = _coords3d(g_ref)  # bohr
    Q = _coords3d(g_mob)  # bohr
    if P.shape != Q.shape:
        raise ValueError(f"Different atom counts: {P.shape[0]} vs {Q.shape[0]}")
    N = P.shape[0]
    idx = _freeze_union(g_ref, g_mob, n_atoms=N)

    mode = "kabsch"
    report_all_atoms = False

    # ---- 1 anchor ----
    if len(idx) == 1:
        i = idx[0]
        before = _rmsd(P, Q)  # all-atom RMSD
        p0, q0 = P[i].copy(), Q[i].copy()
        Q_shift = Q + (p0 - q0)                # match the anchor point
        P_rel, Q_rel = P - p0, Q_shift - p0
        U, _, Vt = np.linalg.svd(P_rel.T @ Q_rel)
        R = Vt.T @ U.T
        if np.linalg.det(R) < 0.0:
            Vt[-1] *= -1.0
            R = Vt.T @ U.T
        Q_aln = (Q_rel @ R) + p0
        _set_all_coords_disabling_freeze(g_mob, Q_aln)
        after = _rmsd(P, Q_aln)
        mode = "one_anchor"
        if verbose:
            click.echo(f"[align] one-anchor: RMSD {before:.6f} Å → {after:.6f} Å (idx={i})")
        return {"before_A": before, "after_A": after, "n_used": 1, "mode": mode}

    # ---- 2 anchors ----
    if len(idx) == 2:
        i0, i1 = idx[0], idx[1]
        p0, p1 = P[i0].copy(), P[i1].copy()
        q0, q1 = Q[i0].copy(), Q[i1].copy()
        vP, vQ = p1 - p0, q1 - q0

        # Only use the special two-anchor treatment when the axis is well-defined.
        if np.linalg.norm(vP) > 1e-16 and np.linalg.norm(vQ) > 1e-16:
            before = _rmsd(P, Q)  # all-atom RMSD
            pm, qm = 0.5 * (p0 + p1), 0.5 * (q0 + q1)
            Q0 = Q + (pm - qm)                      # match midpoints
            R_align = _rotation_align_vectors(vQ, vP)
            Q0 = ((Q0 - pm) @ R_align.T) + pm       # align axis direction (right-multiply → .T)

            u = vP / (np.linalg.norm(vP) + 1e-16)
            c = pm
            P_perp = _orth_proj_perp(u)
            A = (P - c) @ P_perp.T
            B = (Q0 - c) @ P_perp.T
            cross_u_B = np.cross(u, B)
            s1 = float(np.sum(A * B))
            s2 = float(np.sum(A * cross_u_B))
            theta = np.arctan2(s2, s1) if (abs(s1) + abs(s2)) > 1e-16 else 0.0

            R_axis = _rodrigues(u, theta)
            Q1 = ((Q0 - c) @ R_axis.T) + c
            _set_all_coords_disabling_freeze(g_mob, Q1)
            after = _rmsd(P, Q1)
            mode = "two_anchor"
            if verbose:
                click.echo(f"[align] two-anchors: RMSD {before:.6f} Å → {after:.6f} Å (idx=({i0},{i1}))")
            return {"before_A": before, "after_A": after, "n_used": 2, "mode": mode}
        # If the axis is degenerate, fall through to the generic Kabsch case.
        report_all_atoms = True

    # ---- Default: Kabsch (selected freeze atoms or all atoms) ----
    if len(idx) > 0:
        use = np.zeros(N, bool)
        for k in idx:
            if 0 <= k < N:
                use[k] = True
    else:
        use = np.ones(N, bool)

    P_sel, Q_sel = P[use], Q[use]
    n_used = int(P_sel.shape[0])

    before_sel = _rmsd(P_sel, Q_sel)
    before_report = _rmsd(P, Q) if report_all_atoms else before_sel

    R, t = kabsch_R_t(P_sel, Q_sel)
    Q_aln = (Q @ R) + t
    _set_all_coords_disabling_freeze(g_mob, Q_aln)

    after_sel = _rmsd(P_sel, Q_aln[use])
    after_report = _rmsd(P, Q_aln) if report_all_atoms else after_sel

    if verbose:
        click.echo(f"[align] kabsch:     RMSD {before_report:.6f} Å → {after_report:.6f} Å (used {n_used})")

    return {"before_A": before_report, "after_A": after_report, "n_used": n_used, "mode": mode}


# =============================================================================
# Scan + relaxation (stepwise matching of union(`freeze_atoms`) to the reference)
# =============================================================================

def scan_freeze_atoms_toward_target_inplace(
    g_ref,
    g_mob,
    *,
    step_A: float = 0.1,
    per_step_cycles: int = 50,
    final_cycles: int = 200,
    max_steps: int = 1000,
    thresh: str = "gau",
    shared_calc=None,
    out_dir: Path = Path("./result_align_refine/"),
    charge: int = 0,
    spin: int = 1,
    model: str = "uma-s-1p1",
    device: str = "auto",
    verbose: bool = True,
) -> Dict[str, Any]:
    """
    Move the **union** of `freeze_atoms` from (`g_ref`, `g_mob`) toward the reference (`g_ref`) by `step_A` Å
    per iteration. At each step, keep the anchors fixed and run a short LBFGS relaxation on the remaining atoms.
    Finally, enforce exact coincidence of the anchors and perform a finishing relaxation. Updates `g_mob` in place.

    Args:
        step_A: Step size in Å for anchor movement toward the reference per iteration.
        per_step_cycles: Maximum LBFGS iterations per intermediate relaxation with anchors fixed.
        final_cycles: Maximum LBFGS iterations for the final relaxation after anchors coincide.
        max_steps: Maximum number of scan steps before giving up.
        thresh: LBFGS convergence threshold keyword (passed to pysisyphus LBFGS).
        shared_calc: Optional calculator to reuse across geometries.
        out_dir: Output directory for LBFGS runs.
        charge, spin, model, device: Calculator configuration for UMA if no calculator is attached.
        verbose: If True, prints progress.

    Returns:
        dict(max_remaining_A, n_steps, converged)
    """
    P = _coords3d(g_ref)  # bohr
    Q = _coords3d(g_mob)  # bohr
    if P.shape != Q.shape:
        raise ValueError(f"Different atom counts: {P.shape[0]} vs {Q.shape[0]}")
    N = P.shape[0]

    original_freeze = np.array(getattr(g_mob, "freeze_atoms", []), int)
    try:
        idx = _freeze_union(g_ref, g_mob, n_atoms=N)

        if len(idx) == 0:
            if verbose:
                click.echo("[scan] freeze_atoms list is empty. Skipping scan and relaxation.")
            return {"max_remaining_A": 0.0, "n_steps": 0, "converged": True}

        _attach_calc_if_needed(g_mob, shared_calc, charge=charge, spin=spin, model=model, device=device)

        out_dir = Path(out_dir)
        out_dir.mkdir(parents=True, exist_ok=True)

        step_bohr = float(step_A) / BOHR2ANG
        eps = 1e-12
        n_steps_done = 0
        converged = False
        max_remaining_A = None

        for istep in range(1, max_steps + 1):
            Q = _coords3d(g_mob)
            d = P[idx] - Q[idx]                    # bohr
            rem_bohr = np.linalg.norm(d, axis=1)
            max_rem_bohr = float(rem_bohr.max()) if len(rem_bohr) else 0.0
            max_remaining_A = max_rem_bohr * BOHR2ANG
            if verbose:
                click.echo(f"[scan] step {istep:03d}: max remaining = {max_remaining_A:.6f} Å")

            if max_rem_bohr <= step_bohr + 1e-12:
                # Final step: enforce exact coincidence
                Q_new = Q.copy()
                Q_new[idx] = P[idx]
                _set_all_coords_disabling_freeze(g_mob, Q_new)
                try:
                    g_mob.freeze_atoms = np.array(idx, int)
                    LBFGS(
                        g_mob,
                        out_dir=str(out_dir),
                        max_cycles=int(final_cycles),
                        print_every=100,
                        thresh=thresh,
                        dump=False,
                    ).run()
                except (ZeroStepLength, OptimizationError) as e:
                    if verbose:
                        click.echo(f"[scan] WARNING: Exception occurred in final relaxation: {e} (continuing)", err=True)
                g_mob.freeze_atoms = np.array([], int)
                converged = True
                n_steps_done = istep
                break

            # Take one step forward toward the target
            move = np.zeros_like(d)
            sel = rem_bohr > eps
            move[sel] = (d[sel] / rem_bohr[sel, None]) * step_bohr
            Q_next = Q.copy()
            Q_next[idx] = Q[idx] + move

            _set_all_coords_disabling_freeze(g_mob, Q_next)
            try:
                g_mob.freeze_atoms = np.array(idx, int)
                LBFGS(
                    g_mob,
                    out_dir=str(out_dir),
                    max_cycles=int(per_step_cycles),
                    print_every=100,
                    thresh=thresh,
                    dump=False,
                ).run()
            except (ZeroStepLength, OptimizationError) as e:
                if verbose:
                    click.echo(f"[scan] WARNING: Exception occurred in relaxation: {e} (continuing)", err=True)
            finally:
                g_mob.freeze_atoms = np.array([], int)

            n_steps_done = istep
        else:
            if verbose:
                click.echo(f"[scan] WARNING: Reached max_steps={max_steps}.", err=True)

        return {"max_remaining_A": float(max_remaining_A or 0.0),
                "n_steps": int(n_steps_done),
                "converged": bool(converged)}
    finally:
        g_mob.freeze_atoms = original_freeze


# =============================================================================
# High-level API: pair / sequence
# =============================================================================

def align_and_refine_pair_inplace(
    g_ref,
    g_mob,
    *,
    shared_calc=None,
    out_dir: Path = Path("./result_align_refine/"),
    step_A: float = 0.1,
    per_step_cycles: int = 50,
    final_cycles: int = 200,
    max_steps: int = 1000,
    thresh: str = "gau",
    charge: int = 0,
    spin: int = 1,
    model: str = "uma-s-1p1",
    device: str = "auto",
    verbose: bool = True,
) -> Dict[str, Any]:
    """
    For a pair (g_ref, g_mob), perform:
      (1) rigid alignment (special cases for union length = 1/2, otherwise Kabsch), then
      (2) scan + relaxation (stepwise matching of the union of `freeze_atoms` to the reference).
    Updates `g_mob` in place.

    Args mirror `scan_freeze_atoms_toward_target_inplace` (including `thresh`).

    Returns:
        {
          "align": {before_A, after_A, n_used, mode},
          "scan":  {max_remaining_A, n_steps, converged}
        }
    """
    align_res = align_second_to_first_kabsch_inplace(g_ref, g_mob, verbose=verbose)
    scan_res = scan_freeze_atoms_toward_target_inplace(
        g_ref, g_mob,
        step_A=step_A,
        per_step_cycles=per_step_cycles,
        final_cycles=final_cycles,
        max_steps=max_steps,
        thresh=thresh,
        shared_calc=shared_calc,
        out_dir=out_dir,
        charge=charge, spin=spin, model=model, device=device,
        verbose=verbose,
    )
    return {"align": align_res, "scan": scan_res}


def align_and_refine_sequence_inplace(
    geoms: Sequence[Any],
    *,
    shared_calc=None,
    out_dir: Path = Path("./result_align_refine/"),
    step_A: float = 0.1,
    per_step_cycles: int = 1000,
    final_cycles: int = 1000,
    max_steps: int = 10000,
    thresh: str = "gau",
    charge: int = 0,
    spin: int = 1,
    model: str = "uma-s-1p1",
    device: str = "auto",
    verbose: bool = True,
) -> List[Dict[str, Any]]:
    """
    For a list [g0, g1, g2, ...], apply `align_and_refine_pair_inplace` in order:
    (g0←g1), (g1←g2), ... i.e., each g_{i+1} is aligned/refined to g_i.
    Returns a list of per-pair result dicts.

    Intended to be used after preoptimization in `path_search.py`.
    """
    geoms = list(geoms)
    if len(geoms) <= 1:
        return []

    out_dir = Path(out_dir)

    results: List[Dict[str, Any]] = []
    for i in range(len(geoms) - 1):
        g_ref = geoms[i]
        g_mob = geoms[i + 1]
        pair_out = out_dir / f"pair_{i:02d}"

        if verbose:
            click.echo(f"[stage] [align+scan] Pair {i:02d}: image {i} (ref) ← image {i+1} (mobile)")

        res = align_and_refine_pair_inplace(
            g_ref, g_mob,
            shared_calc=shared_calc,
            out_dir=pair_out,
            step_A=step_A,
            per_step_cycles=per_step_cycles,
            final_cycles=final_cycles,
            max_steps=max_steps,
            thresh=thresh,
            charge=charge, spin=spin, model=model, device=device,
            verbose=verbose,
        )
        results.append(res)

    return results


__all__ = [
    "align_second_to_first_kabsch_inplace",
    "scan_freeze_atoms_toward_target_inplace",
    "align_and_refine_pair_inplace",
    "align_and_refine_sequence_inplace",
    "kabsch_R_t",
]



from typing import List, Optional
import warnings

import numpy as np
import torch
from numpy.typing import NDArray

from pysisyphus.intcoords.Coords import CoordSys


class CartesianCoords(CoordSys):
    def __init__(
        self,
        atoms,
        coords3d: NDArray,
        masses,
        freeze_atoms=None,
        *,
        mass_weighted=False,
        **kwargs,
    ):
        for key, val in kwargs.items():
            # f-string don't seem to work when pytest reports the warnings after a test
            # run, so we construct to warning before it is issued.
            msg = (
                # Also printing the value is too much
                # f"Keyword '{key}': '{val}' is not supported by this coordinate system!"
                f"Keyword '{key}' is not supported by this coordinate system!"
            )
            warnings.warn(msg)
        self.atoms = atoms
        self._coords3d = coords3d
        self.atom_num = len(self.atoms)
        self.masses = masses
        if freeze_atoms is None:
            freeze_atoms = list()
        self.freeze_atoms = freeze_atoms
        self.mass_weighted = mass_weighted

        self.move_mask = np.full(self.atom_num, True, dtype=bool)
        self.move_mask[self.freeze_atoms] = False
        self.move_mask_rep = np.repeat(self.move_mask, 3)
        self.zero_vec = np.zeros_like(coords3d)

    @property
    def masses(self) -> NDArray:
        return self._masses

    @masses.setter
    def masses(self, masses: NDArray):
        assert len(masses) == self.atom_num
        masses = np.array(masses, dtype=float)
        self._masses = masses
        # Also precalculate masses' square roots for mass-weighting
        self._masses_sqrt = np.sqrt(self.masses)
        # and its inverse
        self._inv_masses_rep_sqrt = 1 / np.repeat(self.masses_sqrt, 3)

    @property
    def masses_sqrt(self) -> NDArray:
        return self._masses_sqrt

    @property
    def inv_masses_rep_sqrt(self) -> NDArray:
        return self._inv_masses_rep_sqrt

    @property
    def coords(self) -> NDArray:
        coords3d = self.coords3d.copy()
        if self.mass_weighted:
            coords3d *= self.masses_sqrt[:, None]
        coords3d = coords3d[self.move_mask]
        return coords3d.flatten()

    def transform_forces(self, cart_forces: NDArray) -> NDArray:
        forces = cart_forces.reshape(-1, 3)
        if self.mass_weighted:
            forces /= self.masses_sqrt[:, None]
        forces = forces[self.move_mask]
        return forces.flatten()

    def transform_hessian(
        self, cart_hessian: NDArray, int_gradient: Optional[NDArray] = None
    ):
        if isinstance(cart_hessian, torch.Tensor):
            if self.mass_weighted:
                inv_masses = torch.tensor(
                    self.inv_masses_rep_sqrt,
                    dtype=cart_hessian.dtype,
                    device=cart_hessian.device,
                )
                M_inv = torch.diag(inv_masses)
                cart_hessian = M_inv @ cart_hessian @ M_inv
            mask = torch.tensor(
                self.move_mask_rep, dtype=torch.bool, device=cart_hessian.device
            )
            cart_hessian = cart_hessian[mask][:, mask]
            return cart_hessian
        if self.mass_weighted:
            M_inv = np.diag(self.inv_masses_rep_sqrt)
            cart_hessian = M_inv @ cart_hessian @ M_inv
        cart_hessian = cart_hessian[self.move_mask_rep][:, self.move_mask_rep]
        return cart_hessian

    def transform_int_step(
        self,
        step: NDArray,
        update_constraints: Optional[bool] = False,
        pure: Optional[bool] = False,
    ) -> NDArray:
        if update_constraints:
            raise Exception("update_constraints is currently ignored!")
        full_step = self.zero_vec.copy()
        full_step[self.move_mask] = step.reshape(-1, 3)
        if self.mass_weighted:
            full_step /= self.masses_sqrt[:, None]
        if not pure:
            self.coords3d += full_step
        return full_step.flatten()

    def project_hessian(self, hessian):
        return hessian

    @property
    def coords3d(self) -> NDArray:
        return self._coords3d

    @coords3d.setter
    def coords3d(self, coords3d: NDArray):
        self._coords3d = coords3d.reshape(-1, 3)

    @property
    def typed_prims(self) -> List:
        return list()


class MWCartesianCoords(CartesianCoords):
    def __init__(self, *args, **kwargs):
        kwargs["mass_weighted"] = True
        super().__init__(*args, **kwargs)
